/**
 * Copyright (c) 2009 Sonatype, Inc. All rights reserved.
 *
 * This program is licensed to you under the Apache License Version 2.0,
 * and you may not use this file except in compliance with the Apache License Version 2.0.
 * You may obtain a copy of the Apache License Version 2.0 at http://www.apache.org/licenses/LICENSE-2.0.
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the Apache License Version 2.0 is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Apache License Version 2.0 for the specific language governing permissions and limitations there under.
 */
package org.codehaus.plexus;

import java.util.ArrayList;
import java.util.List;

import org.codehaus.plexus.context.Context;
import org.codehaus.plexus.context.ContextException;
import org.codehaus.plexus.logging.LogEnabled;
import org.codehaus.plexus.logging.Logger;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.Contextualizable;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.Disposable;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.Initializable;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.InitializationException;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.Startable;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.StartingException;
import org.sonatype.guice.bean.inject.PropertyBinding;
import org.sonatype.guice.bean.reflect.BeanProperty;
import org.sonatype.guice.plexus.binders.PlexusBeanManager;

import com.google.inject.ProvisionException;

/**
 * {@link PlexusBeanManager} that manages Plexus components requiring lifecycle management.
 */
final class PlexusLifecycleManager
    implements PlexusBeanManager
{
    // ----------------------------------------------------------------------
    // Implementation fields
    // ----------------------------------------------------------------------

    private final List<Startable> startableBeans = new ArrayList<Startable>();

    private final List<Disposable> disposableBeans = new ArrayList<Disposable>();

    private final PlexusContainer container;

    private final Context context;

    // ----------------------------------------------------------------------
    // Constructors
    // ----------------------------------------------------------------------

    PlexusLifecycleManager( final PlexusContainer container, final Context context )
    {
        this.container = container;
        this.context = context;
    }

    // ----------------------------------------------------------------------
    // Public methods
    // ----------------------------------------------------------------------

    public boolean manage( final Class<?> clazz )
    {
        return LogEnabled.class.isAssignableFrom( clazz ) || Contextualizable.class.isAssignableFrom( clazz )
            || Initializable.class.isAssignableFrom( clazz ) || Startable.class.isAssignableFrom( clazz )
            || Disposable.class.isAssignableFrom( clazz );
    }

    @SuppressWarnings( "rawtypes" )
    public PropertyBinding manage( final BeanProperty property )
    {
        final Class clazz = property.getType().getRawType();
        if ( "org.slf4j.Logger".equals( clazz.getName() ) )
        {
            return new PropertyBinding()
            {
                @SuppressWarnings( "unchecked" )
                public <B> void injectProperty( final B bean )
                {
                    property.set( bean, org.slf4j.LoggerFactory.getLogger( bean.getClass() ) );
                }
            };
        }
        if ( Logger.class.equals( clazz ) )
        {
            return new PropertyBinding()
            {
                @SuppressWarnings( "unchecked" )
                public <B> void injectProperty( final B bean )
                {
                    property.set( bean, getPlexusLogger( bean ) );
                }
            };
        }
        return null;
    }

    public synchronized boolean manage( final Object bean )
    {
        /*
         * Record this information before starting, in case there's a problem
         */
        if ( bean instanceof Startable )
        {
            startableBeans.add( (Startable) bean );
        }
        if ( bean instanceof Disposable )
        {
            disposableBeans.add( (Disposable) bean );
        }
        /*
         * Run through the startup phase of the standard plexus "personality"
         */
        if ( bean instanceof LogEnabled )
        {
            ( (LogEnabled) bean ).enableLogging( getPlexusLogger( bean ) );
        }
        if ( bean instanceof Contextualizable )
        {
            contextualize( (Contextualizable) bean );
        }
        if ( bean instanceof Initializable )
        {
            initialize( (Initializable) bean );
        }
        if ( bean instanceof Startable )
        {
            start( (Startable) bean );
        }
        return true;
    }

    public synchronized boolean unmanage( final Object bean )
    {
        if ( startableBeans.remove( bean ) )
        {
            stop( (Startable) bean );
        }
        if ( disposableBeans.remove( bean ) )
        {
            dispose( (Disposable) bean );
        }
        return true;
    }

    public synchronized boolean unmanage()
    {
        // unmanage beans in reverse order
        while ( !startableBeans.isEmpty() )
        {
            stop( startableBeans.remove( startableBeans.size() - 1 ) );
        }
        while ( !disposableBeans.isEmpty() )
        {
            dispose( disposableBeans.remove( disposableBeans.size() - 1 ) );
        }
        return true;
    }

    public PlexusBeanManager manageChild()
    {
        return this;
    }

    // ----------------------------------------------------------------------
    // Shared implementation methods
    // ----------------------------------------------------------------------

    Logger getPlexusLogger( final Object bean )
    {
        return container.getLoggerManager().getLoggerForComponent( bean.getClass().getName(), null );
    }

    // ----------------------------------------------------------------------
    // Implementation methods
    // ----------------------------------------------------------------------

    private void contextualize( final Contextualizable bean )
    {
        try
        {
            bean.contextualize( context );
        }
        catch ( final ContextException e )
        {
            throw new ProvisionException( "Problem contextualizing: " + bean, e );
        }
    }

    private static void initialize( final Initializable bean )
    {
        try
        {
            bean.initialize();
        }
        catch ( final InitializationException e )
        {
            throw new ProvisionException( "Problem initializing: " + bean, e );
        }
    }

    private static void start( final Startable bean )
    {
        try
        {
            bean.start();
        }
        catch ( final StartingException e )
        {
            throw new ProvisionException( "Problem starting: " + bean, e );
        }
    }

    private void stop( final Startable bean )
    {
        try
        {
            bean.stop();
        }
        catch ( final Throwable e )
        {
            error( "Problem stopping: " + bean.getClass(), e );
        }
    }

    private void dispose( final Disposable bean )
    {
        try
        {
            bean.dispose();
        }
        catch ( final Throwable e )
        {
            error( "Problem disposing: " + bean.getClass(), e );
        }
    }

    private void error( final String message, final Throwable cause )
    {
        try
        {
            container.getLogger().error( message, cause );
        }
        catch ( final Throwable ignore )
        {
            System.err.println( message );
            cause.printStackTrace();
        }
    }
}
